#!/bin/env python

from __future__ import print_function
from qds_sdk.qubole import Qubole
from qds_sdk.commands import *
from qds_sdk.cluster import *
import qds_sdk.exception
from qds_sdk.scheduler import SchedulerCmdLine
from qds_sdk.actions import ActionCmdLine
from qds_sdk.report import ReportCmdLine
from qds_sdk.dbtaps import DbTapCmdLine
from qds_sdk.role import RoleCmdLine
from qds_sdk.group import GroupCmdLine
from qds_sdk.account import AccountCmdLine
from qds_sdk.app import AppCmdLine
from qds_sdk.nezha import NezhaCmdLine
from qds_sdk.user import UserCmdLine
from qds_sdk.template import TemplateCmdLine

import os
import sys
import traceback
import logging
import json
from optparse import OptionParser

log = logging.getLogger("qds")
CommandClasses = {
    "hivecmd": HiveCommand,
    "sparkcmd": SparkCommand,
    "dbtapquerycmd": DbTapQueryCommand,
    "pigcmd":  PigCommand,
    "hadoopcmd": HadoopCommand,
    "shellcmd": ShellCommand,
    "dbexportcmd": DbExportCommand,
    "dbimportcmd": DbImportCommand,
    "prestocmd": PrestoCommand
}

usage_str = (
    "Usage: qds.py [options] <subcommand>\n"
    "\nCommand subcommands:\n"
    "  <hivecmd|hadoopcmd|prestocmd|pigcmd|shellcmd|dbexportcmd|dbimportcmd|dbtapquerycmd|sparkcmd> <action>\n"
    "    submit [cmd-specific-args .. ] : submit cmd & print id\n"
    "    run [cmd-specific-args .. ] : submit cmd & wait. print results\n"
    "    check <id> : print the cmd object for this id\n"
    "    cancel <id> : cancels the cmd with this id\n"
    "    getresult <id> : get the results for the cmd with this id\n"
    "    getlog <id> : get the logs for the cmd with this id\n"
    "\nCluster subcommand:\n"
    "  cluster <action>\n"
    "    create: create a new cluster\n"
    "    delete: delete an existing cluster\n"
    "    update: update the settings of an existing cluster\n"
    "    clone: clone a cluster from an existing one\n"
    "    list: list existing cluster(s)\n"
    "    start: start an existing cluster\n"
    "    terminate: terminate a running cluster\n"
    "    status: show whether the cluster is up or down\n"
    "    reassign_label: reassign label from one cluster to another\n"
    "    snapshot: take snapshot of cluster\n"
    "    restore_point: restore cluster from snapshot\n"
    "    add_node: add a node to existing cluster\n"
    "    remove_node: remove a node to existing cluster\n"
    "    update_node: update a node on a existing cluster\n"
    "    get_snapshot_schedule: get details of scheduled snapshots on a hbase cluster\n"
    "    update_snapshot_schedule: update scheduled snapshots on a hbase cluster\n"
    "\nDbTap subcommand:\n"
    "  dbtap --help\n"
    "\nReport subcommand:\n"
    "  report --help\n"
    "\nGroup subcommand:\n"
    "  group --help\n"
    "\nRole subcommand:\n"
    "  role --help\n"
    "\nApp subcommand:\n"
    "  app --help\n"
    "\nAction subcommand:\n"
    "  action --help\n"
    "\nScheduler subcommand:\n"
    "  scheduler --help\n"
    "\nTemplate subcommand:\n"
    "  template --help\n"
    "\nAccount subcommand:\n"
    "  account --help\n"
    "\nNezha subcommand:\n"
    "  nezha --help\n"
    "\nUser subcommad:\n"
    "  user --help\n")


def usage(parser=None):
    if parser is None:
        sys.stderr.write(usage_str)
    else:
        parser.print_help()
    sys.exit(1)


def checkargs_id(args):
    if len(args) != 1:
        sys.stderr.write("expecting single argument command id\n")
        usage()


def submitaction(cmdclass, args):
    args = cmdclass.parse(args)
    if args is not None:
        args.pop("print_logs") # This is only useful while using the 'run' action.
        args.pop("print_logs_live") # This is only useful while using the 'run' action.
        cmd = cmdclass.create(**args)
        print("Submitted %s, Id: %s" % (cmdclass.__name__, cmd.id))
        return 0


def _getresult(cmdclass, cmd):
    if Command.is_success(cmd.status):
        log.info("Fetching results for %s, Id: %s" % (cmdclass.__name__, cmd.id))
        cmd.get_results(sys.stdout, delim='\t')
        return 0
    else:
        log.error("Cannot fetch results - command Id: %s failed with status: %s" % (cmd.id, cmd.status))
        return 1


def runaction(cmdclass, args):
    args = cmdclass.parse(args)
    if args is not None:
        print_logs = args.pop("print_logs") # We don't want to send this to the API.
        cmd = cmdclass.run(**args)
        if print_logs:
            sys.stderr.write(cmd.get_log())
        return _getresult(cmdclass, cmd)


def checkaction(cmdclass, args):
    checkargs_id(args)
    o = cmdclass.find(args.pop(0))
    print(str(o))
    return 0


def cancelaction(cmdclass, args):
    checkargs_id(args)
    r = cmdclass.cancel_id(args.pop(0))
    skey = 'kill_succeeded'
    if r.get(skey) is None:
        sys.stderr.write("Invalid Json Response %s - missing field '%s'" % (str(r), skey))
        return 11
    elif r['kill_succeeded']:
        print("Command killed successfully")
        return 0
    else:
        sys.stderr.write("Cancel failed with reason '%s'\n" % r.get('result'))
        return 12


def getresultaction(cmdclass, args):
    checkargs_id(args)
    cmd = cmdclass.find(args.pop(0))
    return _getresult(cmdclass, cmd)


def getlogaction(cmdclass, args):
    checkargs_id(args)
    print(cmdclass.get_log_id(args.pop(0)))
    return 0


def getjobsaction(cmdclass, args):
    checkargs_id(args)
    cmd = cmdclass.find(args.pop(0))
    if Command.is_done(cmd.status):
        log.info("Fetching jobs for %s, Id: %s" % (cmdclass.__name__, cmd.id))
        print(cmdclass.get_jobs_id(cmd.id))
        return 0
    else:
        log.error("Cannot fetch jobs - command Id: %s is not done. Status: %s" % (cmd.id, cmd.status))
        return 1


def cmdmain(cmd, args):
    cmdclass = CommandClasses[cmd]

    actionset = set(["submit", "run", "check", "cancel", "getresult", "getlog", "getjobs"])
    if len(args) < 1:
        sys.stderr.write("missing argument containing action\n")
        usage()

    action = args.pop(0)
    if action not in actionset:
        sys.stderr.write("action must be one of <%s>\n" % "|".join(actionset))
        usage()

    return globals()[action + "action"](cmdclass, args)


def checkargs_cluster_id_label(args):
    if len(args) != 1:
        sys.stderr.write("expecting single argument cluster id or cluster label\n")
        usage()


def cluster_create_action(clusterclass, args, api_version=1.2):
    arguments = clusterclass._parse_create_update(args, "create", api_version)
    cluster_info = _create_cluster_info(arguments, api_version)
    result = clusterclass.create(cluster_info.minimal_payload())
    print(json.dumps(result, indent=4))
    return 0


def cluster_update_action(clusterclass, args, api_version=1.2):
    arguments = clusterclass._parse_create_update(args, "update", api_version)
    cluster_info = _create_cluster_info(arguments, api_version)
    result = clusterclass.update(arguments.cluster_id_label, cluster_info.minimal_payload())
    print(json.dumps(result, indent=4))
    return 0

def cluster_clone_action(clusterclass, args, api_version=1.2):
    arguments = clusterclass._parse_create_update(args, "clone", api_version)
    cluster_info = _create_cluster_info(arguments, api_version)
    result = clusterclass.clone(arguments.cluster_id_label, cluster_info.minimal_payload())
    print(json.dumps(result, indent=4))
    return 0

def _create_cluster_info(arguments, api_version):
    custom_config = _read_file(arguments.custom_config_file, "custom config file")
    presto_custom_config = _read_file(arguments.presto_custom_config_file, "presto custom config file")
    fairscheduler_config_xml = _read_file(arguments.fairscheduler_config_xml_file, "config xml file")
    customer_ssh_key = _read_file(arguments.customer_ssh_key_file, "customer ssh key file")

    cluster_info = None
    if api_version >= 1.3:
        cluster_info = ClusterInfoV13(arguments.label, api_version)
        cluster_info.set_cluster_info(aws_access_key_id=arguments.aws_access_key_id,
                                      aws_secret_access_key=arguments.aws_secret_access_key,
                                      aws_region=arguments.aws_region,
                                      aws_availability_zone=arguments.aws_availability_zone,
                                      vpc_id=arguments.vpc_id,
                                      subnet_id=arguments.subnet_id,
                                      master_elastic_ip=arguments.master_elastic_ip,
                                      disallow_cluster_termination=arguments.disallow_cluster_termination,
                                      enable_ganglia_monitoring=arguments.enable_ganglia_monitoring,
                                      node_bootstrap_file=arguments.node_bootstrap_file,
                                      master_instance_type=arguments.master_instance_type,
                                      slave_instance_type=arguments.slave_instance_type,
                                      initial_nodes=arguments.initial_nodes,
                                      max_nodes=arguments.max_nodes,
                                      slave_request_type=arguments.slave_request_type,
                                      fallback_to_ondemand=arguments.fallback_to_ondemand,
                                      custom_config=custom_config,
                                      use_hbase=arguments.use_hbase,
                                      custom_ec2_tags=arguments.custom_ec2_tags,
                                      use_hadoop2=arguments.use_hadoop2,
                                      use_spark=arguments.use_spark,
                                      use_qubole_placement_policy=arguments.use_qubole_placement_policy,
                                      maximum_bid_price_percentage=arguments.maximum_bid_price_percentage,
                                      timeout_for_request=arguments.timeout_for_request,
                                      maximum_spot_instance_percentage=arguments.maximum_spot_instance_percentage,
                                      stable_maximum_bid_price_percentage=arguments.stable_maximum_bid_price_percentage,
                                      stable_timeout_for_request=arguments.stable_timeout_for_request,
                                      stable_allow_fallback=arguments.stable_allow_fallback,
                                      ebs_volume_count=arguments.ebs_volume_count,
                                      ebs_volume_type=arguments.ebs_volume_type,
                                      ebs_volume_size=arguments.ebs_volume_size,
                                      fairscheduler_config_xml=fairscheduler_config_xml,
                                      default_pool=arguments.default_pool,
                                      encrypted_ephemerals=arguments.encrypted_ephemerals,
                                      ssh_public_key=customer_ssh_key,
                                      persistent_security_group=arguments.persistent_security_group,
                                      enable_presto=arguments.enable_presto,
                                      bastion_node_public_dns=arguments.bastion_node_public_dns,
                                      role_instance_profile=arguments.role_instance_profile,
                                      presto_custom_config=presto_custom_config,
                                      is_ha=arguments.is_ha)
    else:
        cluster_info = ClusterInfo(arguments.label,
                                   arguments.aws_access_key_id,
                                   arguments.aws_secret_access_key,
                                   arguments.disallow_cluster_termination,
                                   arguments.enable_ganglia_monitoring,
                                   arguments.node_bootstrap_file,)

        cluster_info.set_ec2_settings(arguments.aws_region,
                                      arguments.aws_availability_zone,
                                      arguments.vpc_id,
                                      arguments.subnet_id,
                                      arguments.master_elastic_ip,
                                      arguments.role_instance_profile,
                                      arguments.bastion_node_public_dns)

        cluster_info.set_hadoop_settings(arguments.master_instance_type,
                                         arguments.slave_instance_type,
                                         arguments.initial_nodes,
                                         arguments.max_nodes,
                                         custom_config,
                                         arguments.slave_request_type,
                                         arguments.use_hbase,
                                         arguments.custom_ec2_tags,
                                         arguments.use_hadoop2,
                                         arguments.use_spark,
                                         arguments.is_ha)

        cluster_info.set_spot_instance_settings(
              arguments.maximum_bid_price_percentage,
              arguments.timeout_for_request,
              arguments.maximum_spot_instance_percentage)

        cluster_info.set_stable_spot_instance_settings(
              arguments.stable_maximum_bid_price_percentage,
              arguments.stable_timeout_for_request,
              arguments.stable_allow_fallback)

        cluster_info.set_fairscheduler_settings(fairscheduler_config_xml,
                                            arguments.default_pool)

        cluster_info.set_security_settings(arguments.encrypted_ephemerals,
                                           customer_ssh_key,
                                           arguments.persistent_security_group)

        cluster_info.set_presto_settings(arguments.enable_presto,
                                         presto_custom_config)

    return cluster_info

def _read_file(file_path, file_name):
    file_content = None
    if file_path is not None:
        try:
            file_content = open(file_path).read()
        except IOError as e:
            sys.stderr.write("Unable to read %s: %s\n" % (file_name, str(e)))
            usage()
    return file_content

def cluster_delete_action(clusterclass, args):
    checkargs_cluster_id_label(args)
    result = clusterclass.delete(args.pop(0))
    print(json.dumps(result, indent=4))
    return 0


def cluster_list_action(clusterclass, args):
    arguments = clusterclass._parse_list(args)
    if arguments['cluster_id'] is not None:
        result = clusterclass.show(arguments['cluster_id'])
    elif arguments['label'] is not None:
        result = clusterclass.show(arguments['label'])
    elif arguments['state'] is not None:
        result = clusterclass.list(state=arguments['state'])
    else:
        result = clusterclass.list()
    print(json.dumps(result, indent=4))
    return 0


def cluster_start_action(clusterclass, args):
    checkargs_cluster_id_label(args)
    result = clusterclass.start(args.pop(0))
    print(json.dumps(result, indent=4))
    return 0


def cluster_terminate_action(clusterclass, args):
    checkargs_cluster_id_label(args)
    result = clusterclass.terminate(args.pop(0))
    print(json.dumps(result, indent=4))
    return 0


def cluster_status_action(clusterclass, args):
    checkargs_cluster_id_label(args)
    result = clusterclass.status(args.pop(0))
    print(json.dumps(result, indent=4))
    return 0


def cluster_reassign_label_action(clusterclass, args):
    arguments = clusterclass._parse_reassign_label(args)
    result = clusterclass.reassign_label(arguments.destination_cluster,
                                         arguments.label)
    print(json.dumps(result, indent=4))
    return 0

def cluster_snapshot_action(clusterclass, args):
    arguments = clusterclass._parse_snapshot_restore_command(args, "snapshot")
    result = clusterclass.snapshot(arguments.cluster_id or arguments.label, arguments.s3_location, arguments.backup_type)
    print(json.dumps(result, indent=4))
    return 0

def cluster_restore_point_action(clusterclass, args):
    arguments = clusterclass._parse_snapshot_restore_command(args, "restore_point")
    result = clusterclass.restore_point(arguments.cluster_id or arguments.label, arguments.s3_location, arguments.backup_id, arguments.table_names, arguments.no_overwrite, arguments.no_automatic)
    print(json.dumps(result, indent=4))
    return 0

def cluster_get_snapshot_schedule_action(clusterclass, args):
    arguments = clusterclass._parse_get_snapshot_schedule(args)
    result = clusterclass.get_snapshot_schedule(arguments.cluster_id or arguments.label)
    print(json.dumps(result, indent=4))
    return 0

def cluster_update_snapshot_schedule_action(clusterclass, args):
    arguments = clusterclass._parse_update_snapshot_schedule(args)
    result = clusterclass.update_snapshot_schedule(arguments.cluster_id or arguments.label, arguments.s3_location, arguments.frequency_unit, arguments.frequency_num, arguments.status)
    print(json.dumps(result, indent=4))
    return 0

def cluster_add_node_action(clusterclass, args):
    arguments = clusterclass._parse_cluster_manage_command(args, action = "add")
    result = clusterclass.add_node(arguments.cluster_id or arguments.label)
    print(json.dumps(result, indent=4))
    return 0

def cluster_remove_node_action(clusterclass, args):
    arguments = clusterclass._parse_cluster_manage_command(args, action = "remove")
    result = clusterclass.remove_node(arguments.cluster_id or arguments.label, arguments.private_dns)
    print(json.dumps(result, indent=4))
    return 0

def cluster_update_node_action(clusterclass, args):
    arguments = clusterclass._parse_cluster_manage_command(args, action = "update")
    result = clusterclass.update_node(arguments.cluster_id or arguments.label, arguments.command, arguments.private_dns)
    print(json.dumps(result, indent=4))
    return 0

def clustermain(args, api_version):
    clusterclass = Cluster
    actionset = set(["create", "delete", "update", "clone", "list", "start", "terminate", "status", "reassign_label", "add_node", "remove_node", "update_node", "snapshot", "restore_point", "get_snapshot_schedule", "update_snapshot_schedule"])

    if len(args) < 1:
        sys.stderr.write("missing argument containing action\n")
        usage()

    action = args.pop(0)
    if action not in actionset:
        sys.stderr.write("action must be one of <%s>\n" % "|".join(actionset))
        usage()
    elif action in set(["create", "update", "clone"]):
        return globals()["cluster_" + action + "_action"](clusterclass, args, api_version)
    else:
        return globals()["cluster_" + action + "_action"](clusterclass, args)

def accountmain(args):
    result = AccountCmdLine.run(args)
    print(result)

def usermain(args):
    result = UserCmdLine.run(args)
    print(result)

def reportmain(args):
    result = ReportCmdLine.run(args)
    print(result)


def actionmain(args):
    result = ActionCmdLine.run(args)
    print(result)

def schedulermain(args):
    result = SchedulerCmdLine.run(args)
    print(result)

def dbtapmain(args):
    result = DbTapCmdLine.run(args)
    print(result)

def rolemain(args):
    result = RoleCmdLine.run(args)
    print(result)

def groupmain(args):
    result = GroupCmdLine.run(args)
    print(result)

def appmain(args):
    result = AppCmdLine.run(args)
    print(result)

def nezhamain(args):
    result = NezhaCmdLine.run(args)
    print(result)

def templatemain(args):
    result = TemplateCmdLine.run(args)
    print(result)
    

def main():
    optparser = OptionParser(usage=usage_str)
    optparser.add_option("--token", dest="api_token",
                         default=os.getenv('QDS_API_TOKEN'),
                         help="api token for accessing Qubole. must be specified via command line or passed in via environment variable QDS_API_TOKEN")

    optparser.add_option("--url", dest="api_url",
                         default=os.getenv('QDS_API_URL'),
                         help="base url for QDS REST API. defaults to https://api.qubole.com/api ")

    optparser.add_option("--version", dest="api_version",
                         default=os.getenv('QDS_API_VERSION'),
                         help="version of REST API to access. defaults to v1.2")

    optparser.add_option("--poll_interval", dest="poll_interval",
                         type=int,
                         default=os.getenv('QDS_POLL_INTERVAL'),
                         help="interval for polling API for completion and other events. defaults to 5s")

    optparser.add_option("--skip_ssl_cert_check", dest="skip_ssl_cert_check", action="store_true",
                         default=False,
                         help="skip verification of server SSL certificate. Insecure: use with caution.")

    optparser.add_option("-v", dest="verbose", action="store_true",
                         default=False,
                         help="verbose mode - info level logging")

    optparser.add_option("--vv", dest="chatty", action="store_true",
                         default=False,
                         help="very verbose mode - debug level logging")

    optparser.disable_interspersed_args()
    (options, args) = optparser.parse_args()
    
    if options.chatty:
        logging.basicConfig(level=logging.DEBUG)
    elif options.verbose:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARN)

    if options.api_token is None:
        sys.stderr.write("No API Token provided\n")
        usage(optparser)

    if options.api_url is None:
        options.api_url = "https://api.qubole.com/api/"

    if options.api_version is None:
        options.api_version = "v1.2"

    if options.poll_interval is None:
        options.poll_interval = 5

    if options.skip_ssl_cert_check is None:
        options.skip_ssl_cert_check = False
    elif options.skip_ssl_cert_check:
        log.warn("Insecure mode enabled: skipping SSL cert verification\n")

    Qubole.configure(api_token=options.api_token,
                     api_url=options.api_url,
                     version=options.api_version,
                     poll_interval=options.poll_interval,
                     skip_ssl_cert_check=options.skip_ssl_cert_check)

    if len(args) < 1:
        sys.stderr.write("Missing first argument containing subcommand\n")
        usage(optparser)

    a0 = args.pop(0)
    if a0 in CommandClasses:
        return cmdmain(a0, args)

    if a0 == "account":
        return accountmain(args)

    if a0 == "cluster":
        api_version_number = float(options.api_version[1:])
        return clustermain(args, api_version_number)

    if a0 == "action":
        return actionmain(args)

    if a0 == "scheduler":
        return schedulermain(args)

    if a0 == "report":
        return reportmain(args)

    if a0 == "dbtap":
        return dbtapmain(args)

    if a0 == "group":
        return groupmain(args)

    if a0 == "role":
        return rolemain(args)

    if a0 == "app":
        return appmain(args)

    if a0 == "nezha":
        return nezhamain(args)

    if a0 == "user":
        return usermain(args)
    if a0 == "template":
        return templatemain(args)

    cmdset = set(CommandClasses.keys())
    sys.stderr.write("First command must be one of <%s>\n" %
                     "|".join(cmdset.union(["cluster", "action", "scheduler", "report",
                       "dbtap", "role", "group", "app", "account", "nezha", "user", "template"])))
    usage(optparser)


if __name__ == '__main__':
    try:
        sys.exit(main())
    except qds_sdk.exception.Error as e:
        sys.stderr.write("Error: Status code %s (%s) from url %s\n" %
                         (e.request.status_code, e.__class__.__name__,
                          e.request.url))
        sys.exit(1)
    except qds_sdk.exception.ParseError as e:
        sys.stderr.write("Error: %s\n" % str(e))
        sys.stderr.write("Usage: %s\n" % e.usage)
        sys.exit(2)
    except Exception:
        traceback.print_exc(file=sys.stderr)
        sys.exit(3)



"""
The cluster module contains the definitions for retrieving and manipulating
cluster information.
"""

from qds_sdk.qubole import Qubole
from qds_sdk.resource import Resource
from argparse import ArgumentParser

import logging
import json

log = logging.getLogger("qds_cluster")


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


class Cluster(Resource):
    """
    qds_sdk.Cluster is the class for retrieving and manipulating cluster
    information.
    """

    rest_entity_path = "clusters"
    api_version = "v1.2"

    @classmethod
    def _parse_list(cls, args):
        """
        Parse command line arguments to construct a dictionary of cluster
        parameters that can be used to determine which clusters to list.

        Args:
            `args`: sequence of arguments

        Returns:
            Dictionary that can be used to determine which clusters to list
        """
        argparser = ArgumentParser(prog="cluster list")

        group = argparser.add_mutually_exclusive_group()

        group.add_argument("--id", dest="cluster_id",
                           help="show cluster with this id")

        group.add_argument("--label", dest="label",
                           help="show cluster with this label")

        group.add_argument("--state", dest="state", action="store",
                           choices=['up', 'down', 'pending', 'terminating'],
                           help="list only clusters in the given state")

        arguments = argparser.parse_args(args)
        return vars(arguments)

    @classmethod
    def list(cls, state=None):
        """
        List existing clusters present in your account.

        Kwargs:
            `state`: list only those clusters which are in this state

        Returns:
            List of clusters satisfying the given criteria
        """
        conn = Qubole.agent()
        if state is None:
            return conn.get(cls.rest_entity_path)
        elif state is not None:
            cluster_list = conn.get(cls.rest_entity_path)
            result = []
            for cluster in cluster_list:
                if state.lower() == cluster['cluster']['state'].lower():
                    result.append(cluster)
            return result

    @classmethod
    def show(cls, cluster_id_label):
        """
        Show information about the cluster with id/label `cluster_id_label`.
        """
        conn = Qubole.agent()
        return conn.get(cls.element_path(cluster_id_label))

    @classmethod
    def status(cls, cluster_id_label):
        """
        Show the status of the cluster with id/label `cluster_id_label`.
        """
        conn = Qubole.agent()
        return conn.get(cls.element_path(cluster_id_label) + "/state")

    @classmethod
    def start(cls, cluster_id_label):
        """
        Start the cluster with id/label `cluster_id_label`.
        """
        conn = Qubole.agent()
        data = {"state": "start"}
        return conn.put(cls.element_path(cluster_id_label) + "/state", data)

    @classmethod
    def terminate(cls, cluster_id_label):
        """
        Terminate the cluster with id/label `cluster_id_label`.
        """
        conn = Qubole.agent()
        data = {"state": "terminate"}
        return conn.put(cls.element_path(cluster_id_label) + "/state", data)

    @classmethod
    def _parse_create_update(cls, args, action, api_version):
        """
        Parse command line arguments to determine cluster parameters that can
        be used to create or update a cluster.

        Args:
            `args`: sequence of arguments

            `action`: "create", "update" or "clone"

        Returns:
            Object that contains cluster parameters
        """
        argparser = ArgumentParser(prog="cluster %s" % action)

        create_required = False
        label_required = False

        if action == "create":
            create_required = True
        elif action == "update":
            argparser.add_argument("cluster_id_label",
                                   help="id/label of the cluster to update")
        elif action == "clone":
            argparser.add_argument("cluster_id_label",
                                   help="id/label of the cluster to update")
            label_required = True

        argparser.add_argument("--label", dest="label",
                               nargs="+", required=(create_required or label_required),
                               help="list of labels for the cluster" +
                                    " (atleast one label is required)")

        ec2_group = argparser.add_argument_group("ec2 settings")
        ec2_group.add_argument("--access-key-id",
                               dest="aws_access_key_id",
                               help="access key id for customer's aws" +
                                    " account. This is required while" +
                                    " creating the cluster",)
        ec2_group.add_argument("--secret-access-key",
                               dest="aws_secret_access_key",
                               help="secret access key for customer's aws" +
                                    " account. This is required while" +
                                    " creating the cluster",)
        ec2_group.add_argument("--aws-region",
                               dest="aws_region",
                               choices=["us-east-1", "us-west-2", "ap-northeast-1", "sa-east-1",
                                        "eu-west-1", "ap-southeast-1", "us-west-1"],
                               help="aws region to create the cluster in",)
        ec2_group.add_argument("--aws-availability-zone",
                               dest="aws_availability_zone",
                               help="availability zone to" +
                                    " create the cluster in",)
        ec2_group.add_argument("--subnet-id",
                               dest="subnet_id",
                               help="subnet to create the cluster in",)
        ec2_group.add_argument("--vpc-id",
                               dest="vpc_id",
                               help="vpc to create the cluster in",)
        ec2_group.add_argument("--master-elastic-ip",
                               dest="master_elastic_ip",
                               help="elastic ip to attach to master",)
        ec2_group.add_argument("--bastion-node-public-dns",
                               dest="bastion_node_public_dns",
                               help="public dns name of the bastion node. Required only if cluster is in private subnet of a EC2-VPC",)
        ec2_group.add_argument("--role-instance-profile",
                               dest="role_instance_profile",
                               help="IAM Role instance profile to attach on cluster",)

        hadoop_group = argparser.add_argument_group("hadoop settings")
        node_config_group = argparser.add_argument_group("node configuration") if (api_version >= 1.3) else hadoop_group

        node_config_group.add_argument("--master-instance-type",
                                  dest="master_instance_type",
                                  help="instance type to use for the hadoop" +
                                       " master node",)
        node_config_group.add_argument("--slave-instance-type",
                                  dest="slave_instance_type",
                                  help="instance type to use for the hadoop" +
                                       " slave nodes",)
        node_config_group.add_argument("--initial-nodes",
                                  dest="initial_nodes",
                                  type=int,
                                  help="number of nodes to start the" +
                                       " cluster with",)
        node_config_group.add_argument("--max-nodes",
                                  dest="max_nodes",
                                  type=int,
                                  help="maximum number of nodes the cluster" +
                                       " may be auto-scaled up to")
        node_config_group.add_argument("--slave-request-type",
                                  dest="slave_request_type",
                                  choices=["ondemand", "spot", "hybrid"],
                                  help="purchasing option for slave instaces",)
        hadoop_group.add_argument("--custom-config",
                                  dest="custom_config_file",
                                  help="location of file containg custom" +
                                       " hadoop configuration overrides")
        hadoop_group.add_argument("--use-hbase", dest="use_hbase",
                                  action="store_true", default=None,
                                  help="Use hbase on this cluster",)
        hadoop_group.add_argument("--is-ha", dest="is_ha",
                                  action="store_true", default=None,
                                  help="Enable HA config for cluster")
        if api_version >= 1.3:
          qubole_placement_policy_group = hadoop_group.add_mutually_exclusive_group()
          qubole_placement_policy_group.add_argument("--use-qubole-placement-policy",
                                              dest="use_qubole_placement_policy",
                                              action="store_true",
                                              default=None,
                                              help="Use Qubole Block Placement policy" +
                                                   " for clusters with spot nodes",)
          qubole_placement_policy_group.add_argument("--no-use-qubole-placement-policy",
                                              dest="use_qubole_placement_policy",
                                              action="store_false",
                                              default=None,
                                              help="Do not use Qubole Block Placement policy" +
                                                   " for clusters with spot nodes",)
          fallback_to_ondemand_group = node_config_group.add_mutually_exclusive_group()
          fallback_to_ondemand_group.add_argument("--fallback-to-ondemand",
                                 dest="fallback_to_ondemand",
                                 action="store_true",
                                 default=None,
                                 help="Fallback to on-demand nodes if spot nodes" +
                                 " could not be obtained. Valid only if slave_request_type is spot",)
          fallback_to_ondemand_group.add_argument("--no-fallback-to-ondemand",
                                 dest="fallback_to_ondemand",
                                 action="store_false",
                                 default=None,
                                 help="Dont Fallback to on-demand nodes if spot nodes" +
                                 " could not be obtained. Valid only if slave_request_type is spot",)
          ebs_volume_group = argparser.add_argument_group("ebs volume settings")
          ebs_volume_group.add_argument("--ebs-volume-count",
                                  dest="ebs_volume_count",
                                  type=int,
                                  help="Number of EBS volumes to attach to" +
                                       " each instance of the cluster",)
          ebs_volume_group.add_argument("--ebs-volume-type",
                                  dest="ebs_volume_type",
                                  choices=["standard", "gp2"],
                                  help=" of the EBS volume. Valid values are " +
                                       "'standard' (magnetic) and 'gp2' (ssd).",)
          ebs_volume_group.add_argument("--ebs-volume-size",
                                  dest="ebs_volume_size",
                                  type=int,
                                  help="Size of each EBS volume, in GB",)

        hadoop2 = hadoop_group.add_mutually_exclusive_group()
        hadoop2.add_argument("--use-hadoop2",
                             dest="use_hadoop2",
                             action="store_true",
                             default=None,
                             help="Use hadoop2 instead of hadoop1")
        hadoop2.add_argument("--use-hadoop1",
                             dest="use_hadoop2",
                             action="store_false",
                             default=None,
                             help="Use hadoop1 instead of hadoop2. This is the default.")
        hadoop2.add_argument("--use-spark",
                           dest="use_spark",
                           action="store_true",
                           default=None,
                           help="Turn on spark for this cluster")

        spot_group = argparser.add_argument_group("spot instance settings" +
                    " (valid only when slave-request-type is hybrid or spot)")
        spot_group.add_argument("--maximum-bid-price-percentage",
                                dest="maximum_bid_price_percentage",
                                type=float,
                                help="maximum value to bid for spot instances" +
                                     " expressed as a percentage of the base" +
                                     " price for the slave node instance type",)
        spot_group.add_argument("--timeout-for-spot-request",
                                dest="timeout_for_request",
                                type=int,
                                help="timeout for a spot instance request" +
                                     " unit: minutes")
        spot_group.add_argument("--maximum-spot-instance-percentage",
                                dest="maximum_spot_instance_percentage",
                                type=int,
                                help="maximum percentage of instances that may" +
                                     " be purchased from the aws spot market," +
                                     " valid only when slave-request-type" +
                                     " is 'hybrid'",)

        stable_spot_group = argparser.add_argument_group("stable spot instance settings")
        stable_spot_group.add_argument("--stable-maximum-bid-price-percentage",
                                       dest="stable_maximum_bid_price_percentage",
                                       type=float,
                                       help="maximum value to bid for stable node spot instances" +
                                       " expressed as a percentage of the base" +
                                       " price for the master and slave node instance types",)
        stable_spot_group.add_argument("--stable-timeout-for-spot-request",
                                       dest="stable_timeout_for_request",
                                       type=int,
                                       help="timeout for a stable node spot instance request" +
                                       " unit: minutes")
        stable_spot_group.add_argument("--stable-allow-fallback",
                                       dest="stable_allow_fallback", default=None,
                                       type=str2bool,
                                       help="whether to fallback to on-demand instances for stable nodes" +
                                       " if spot instances aren't available")

        fairscheduler_group = argparser.add_argument_group(
                              "fairscheduler configuration options")
        fairscheduler_group.add_argument("--fairscheduler-config-xml",
                                         dest="fairscheduler_config_xml_file",
                                         help="location for file containing" +
                                              " xml with custom configuration" +
                                              " for the fairscheduler",)
        fairscheduler_group.add_argument("--fairscheduler-default-pool",
                                         dest="default_pool",
                                         help="default pool for the" +
                                              " fairscheduler",)

        security_group = argparser.add_argument_group("security setttings")
        ephemerals = security_group.add_mutually_exclusive_group()
        ephemerals.add_argument("--encrypted-ephemerals",
                                 dest="encrypted_ephemerals",
                                 action="store_true",
                                 default=None,
                                 help="encrypt the ephemeral drives on" +
                                      " the instance",)
        ephemerals.add_argument("--no-encrypted-ephemerals",
                                 dest="encrypted_ephemerals",
                                 action="store_false",
                                 default=None,
                                 help="don't encrypt the ephemeral drives on" +
                                      " the instance",)

        security_group.add_argument("--customer-ssh-key",
                                    dest="customer_ssh_key_file",
                                    help="location for ssh key to use to" +
                                         " login to the instance")

        security_group.add_argument("--persistent-security-group",
                                    dest="persistent_security_group",
                                    help="a security group to associate with each" +
                                         " node of the cluster. Typically used" +
                                         " to provide access to external hosts")

        presto_group = argparser.add_argument_group("presto settings")
        enabling_presto = presto_group.add_mutually_exclusive_group()
        enabling_presto.add_argument("--enable-presto",
                                  dest="enable_presto",
                                  action="store_true",
                                  default=None,
                                  help="Enable presto for this cluster",)
        enabling_presto.add_argument("--disable-presto",
                                  dest="enable_presto",
                                  action="store_false",
                                  default=None,
                                  help="Disable presto for this cluster",)
        presto_group.add_argument("--presto-custom-config",
                                  dest="presto_custom_config_file",
                                  help="location of file containg custom" +
                                       " presto configuration overrides")

        termination = argparser.add_mutually_exclusive_group()
        termination.add_argument("--disallow-cluster-termination",
                                 dest="disallow_cluster_termination",
                                 action="store_true",
                                 default=None,
                                 help="don't auto-terminate idle clusters," +
                                      " use this with extreme caution",)
        termination.add_argument("--allow-cluster-termination",
                                 dest="disallow_cluster_termination",
                                 action="store_false",
                                 default=None,
                                 help="auto-terminate idle clusters,")

        ganglia = argparser.add_mutually_exclusive_group()
        ganglia.add_argument("--enable-ganglia-monitoring",
                             dest="enable_ganglia_monitoring",
                             action="store_true",
                             default=None,
                             help="enable ganglia monitoring for the" +
                                  " cluster",)
        ganglia.add_argument("--disable-ganglia-monitoring",
                             dest="enable_ganglia_monitoring",
                             action="store_false",
                             default=None,
                             help="disable ganglia monitoring for the" +
                                  " cluster",)

        argparser.add_argument("--node-bootstrap-file",
                dest="node_bootstrap_file",
                help="""name of the node bootstrap file for this cluster. It
                should be in stored in S3 at
                <account-default-location>/scripts/hadoop/NODE_BOOTSTRAP_FILE
                """,)

        argparser.add_argument("--custom-ec2-tags",
                               dest="custom_ec2_tags",
                               help="""Custom ec2 tags to be set on all instances
                               of the cluster. Specified as JSON object (key-value pairs)
                               e.g. --custom-ec2-tags '{"key1":"value1", "key2":"value2"}'
                               """,)

        arguments = argparser.parse_args(args)
        return arguments

    @classmethod
    def create(cls, cluster_info, version=None):
        """
        Create a new cluster using information provided in `cluster_info`.

        Optionally provide the version (eg: v1.3) to use the new version of the
        API. If None we default to v1.2
        """
        conn = Qubole.agent(version=version)
        return conn.post(cls.rest_entity_path, data=cluster_info)

    @classmethod
    def update(cls, cluster_id_label, cluster_info, version=None):
        """
        Update the cluster with id/label `cluster_id_label` using information provided in
        `cluster_info`.

        Optionally provide the version (eg: v1.3) to use the new version of the
        API. If None we default to v1.2
        """
        conn = Qubole.agent(version=version)
        return conn.put(cls.element_path(cluster_id_label), data=cluster_info)

    @classmethod
    def clone(cls, cluster_id_label, cluster_info, version=None):
        """
        Update the cluster with id/label `cluster_id_label` using information provided in
        `cluster_info`.

        Optionally provide the version (eg: v1.3) to use the new version of the
        API. If None we default to v1.2
        """
        conn = Qubole.agent(version=version)
        return conn.post(cls.element_path(cluster_id_label) + '/clone', data=cluster_info)

    @classmethod
    def _parse_cluster_manage_command(cls, args, action):
      """
      Parse command line arguments for cluster manage commands.
      """

      argparser = ArgumentParser(prog="cluster_manage_command")

      group = argparser.add_mutually_exclusive_group(required=True)

      group.add_argument("--id", dest="cluster_id",
                           help="execute on cluster with this id")

      group.add_argument("--label", dest="label",
                           help="execute on cluster with this label")

      if action == "remove" or action == "update":
        argparser.add_argument("--private_dns",
                           help="the private_dns of the machine to be updated/removed", required=True)
      if action == "update":
        argparser.add_argument("--command",
                           help="the update command to be executed", required=True, choices=["replace"])

      arguments = argparser.parse_args(args)
      return arguments

    @classmethod
    def _parse_reassign_label(cls, args):
        """
        Parse command line arguments for reassigning label.
        """
        argparser = ArgumentParser(prog="cluster reassign_label")

        argparser.add_argument("destination_cluster",
                metavar="destination_cluster_id_label",
                help="id/label of the cluster to move the label to")

        argparser.add_argument("label",
                help="label to be moved from the source cluster")

        arguments = argparser.parse_args(args)
        return arguments

    @classmethod
    def reassign_label(cls, destination_cluster, label):
        """
        Reassign a label from one cluster to another.

        Args:
            `destination_cluster`: id/label of the cluster to move the label to

            `label`: label to be moved from the source cluster
        """
        conn = Qubole.agent()
        data = {
                    "destination_cluster": destination_cluster,
                    "label": label
                }
        return conn.put(cls.rest_entity_path + "/reassign-label", data)

    @classmethod
    def delete(cls, cluster_id_label):
        """
        Delete the cluster with id/label `cluster_id_label`.
        """
        conn = Qubole.agent()
        return conn.delete(cls.element_path(cluster_id_label))

    @classmethod
    def _parse_snapshot_restore_command(cls, args, action):
        """
        Parse command line arguments for snapshot command.
        """
        argparser = ArgumentParser(prog="cluster %s" % action)

        group = argparser.add_mutually_exclusive_group(required=True)
        group.add_argument("--id", dest="cluster_id",
                          help="execute on cluster with this id")
        group.add_argument("--label", dest="label",
                          help="execute on cluster with this label")
        argparser.add_argument("--s3_location",
                          help="s3_location where backup is stored", required=True)
        if action == "snapshot":
            argparser.add_argument("--backup_type",
                          help="backup_type: full/incremental, default is full")
        elif action == "restore_point":
            argparser.add_argument("--backup_id",
                          help="back_id from which restoration will be done", required=True)
            argparser.add_argument("--table_names",
                          help="table(s) which are to be restored", required=True)
            argparser.add_argument("--no-overwrite", action="store_false",
                          help="With this option, restore overwrites to the existing table if theres any in restore target")
            argparser.add_argument("--no-automatic", action="store_false",
                          help="With this option, all the dependencies are automatically restored together with this backup image following the correct order")
        arguments = argparser.parse_args(args)

        return arguments

    @classmethod
    def _parse_get_snapshot_schedule(cls, args):
        """
        Parse command line arguments for updating hbase snapshot schedule or to get details.
        """
        argparser = ArgumentParser(prog="cluster snapshot_schedule")

        group = argparser.add_mutually_exclusive_group(required=True)
        group.add_argument("--id", dest="cluster_id",
                          help="execute on cluster with this id")
        group.add_argument("--label", dest="label",
                          help="execute on cluster with this label")
        arguments = argparser.parse_args(args)

        return arguments

    @classmethod
    def _parse_update_snapshot_schedule(cls, args):
        """
        Parse command line arguments for updating hbase snapshot schedule or to get details.
        """
        argparser = ArgumentParser(prog="cluster snapshot_schedule")

        group = argparser.add_mutually_exclusive_group(required=True)
        group.add_argument("--id", dest="cluster_id",
                          help="execute on cluster with this id")
        group.add_argument("--label", dest="label",
                          help="execute on cluster with this label")

        argparser.add_argument("--frequency-num",
                          help="frequency number")
        argparser.add_argument("--frequency-unit",
                          help="frequency unit")
        argparser.add_argument("--s3-location",
                          help="s3_location about where to store snapshots")
        argparser.add_argument("--status",
                          help="status of periodic job you want to change to", choices = ["RUNNING", "SUSPENDED"])

        arguments = argparser.parse_args(args)

        return arguments

    @classmethod
    def snapshot(cls, cluster_id_label, s3_location, backup_type):
        """
        Create hbase snapshot full/incremental
        """
        conn = Qubole.agent()
        parameters = {}
        parameters['s3_location'] = s3_location
        if backup_type:
            parameters['backup_type'] = backup_type
        return conn.post(cls.element_path(cluster_id_label) + "/snapshots", data=parameters)

    @classmethod
    def restore_point(cls, cluster_id_label, s3_location, backup_id, table_names, overwrite=True, automatic=True):
        """
        Restoring cluster from a given hbase snapshot id
        """
        conn = Qubole.agent()
        parameters = {}
        parameters['s3_location'] = s3_location
        parameters['backup_id'] = backup_id
        parameters['table_names'] = table_names
        parameters['overwrite'] = overwrite
        parameters['automatic'] = automatic
        return conn.post(cls.element_path(cluster_id_label) + "/restore_point", data=parameters)

    @classmethod
    def get_snapshot_schedule(cls, cluster_id_label):
        """
        Get details for snapshot schedule
        """
        conn = Qubole.agent()
        return conn.get(cls.element_path(cluster_id_label) + "/snapshot_schedule")

    @classmethod
    def update_snapshot_schedule(cls, cluster_id_label, s3_location=None, frequency_unit=None, frequency_num=None, status=None):
        """
        Update for snapshot schedule
        """
        conn = Qubole.agent()

        data = {}
        if s3_location is not None:
            data["s3_location"] = s3_location
        if frequency_unit is not None:
            data["frequency_unit"] = frequency_unit
        if frequency_num is not None:
            data["frequency_num"] = frequency_num
        if status is not None:
            data["status"] = status
        return conn.put(cls.element_path(cluster_id_label) + "/snapshot_schedule", data)



    @classmethod
    def add_node(cls, cluster_id_label, parameters=None):
      """
      Add a node to an existing cluster
      """
      conn = Qubole.agent()
      parameters = {} if not parameters else parameters
      return conn.post(cls.element_path(cluster_id_label) + "/nodes", data={"parameters" : parameters})

    @classmethod
    def remove_node(cls, cluster_id_label, private_dns, parameters=None):
        """
        Add a node to an existing cluster
        """
        conn = Qubole.agent()
        parameters = {} if not parameters else parameters
        data = {"private_dns" : private_dns, "parameters" : parameters}
        return conn.delete(cls.element_path(cluster_id_label) + "/nodes", data)

    @classmethod
    def update_node(cls, cluster_id_label, command, private_dns, parameters=None):
        """
        Add a node to an existing cluster
        """
        conn = Qubole.agent()
        parameters = {} if not parameters else parameters
        data = {"command" : command, "private_dns" : private_dns, "parameters" : parameters}
        return conn.put(cls.element_path(cluster_id_label) + "/nodes", data)

class ClusterInfo():
    """
    qds_sdk.ClusterInfo is the class which stores information about a cluster.
    You can use objects of this class to create or update a cluster.
    """

    def __init__(self, label, aws_access_key_id, aws_secret_access_key,
                 disallow_cluster_termination=None,
                 enable_ganglia_monitoring=None,
                 node_bootstrap_file=None):
        """
        Args:

        `label`: A list of labels that identify the cluster. At least one label
            must be provided when creating a cluster.

        `aws_access_key_id`: The access key id for customer's aws account. This
            is required for creating the cluster.

        `aws_secret_access_key`: The secret access key for customer's aws
            account. This is required for creating the cluster.

        `disallow_cluster_termination`: Set this to True if you don't want
            qubole to auto-terminate idle clusters. Use this option with
            extreme caution.

        `enable_ganglia_monitoring`: Set this to True if you want to enable
            ganglia monitoring for the cluster.

        `node_bootstrap_file`: name of the node bootstrap file for this
            cluster. It should be in stored in S3 at
            <your-default-location>/scripts/hadoop/
        """
        self.label = label
        self.ec2_settings = {}
        self.ec2_settings['compute_access_key'] = aws_access_key_id
        self.ec2_settings['compute_secret_key'] = aws_secret_access_key
        self.disallow_cluster_termination = disallow_cluster_termination
        self.enable_ganglia_monitoring = enable_ganglia_monitoring
        self.node_bootstrap_file = node_bootstrap_file
        self.hadoop_settings = {}
        self.security_settings = {}
        self.presto_settings = {}

    def set_ec2_settings(self,
                         aws_region=None,
                         aws_availability_zone=None,
                         vpc_id=None,
                         subnet_id=None,
                         master_elastic_ip=None,
                         role_instance_profile=None,
                         bastion_node_public_dns=None):
        """
        Kwargs:

        `aws_region`: AWS region to create the cluster in.

        `aws_availability_zone`: The availability zone to create the cluster
            in.

        `vpc_id`: The vpc to create the cluster in.

        `subnet_id`: The subnet to create the cluster in.

        `bastion_node_public_dns`: Public dns name of the bastion host. Required only if
            cluster is in private subnet.
        """
        self.ec2_settings['aws_region'] = aws_region
        self.ec2_settings['aws_preferred_availability_zone'] = aws_availability_zone
        self.ec2_settings['vpc_id'] = vpc_id
        self.ec2_settings['subnet_id'] = subnet_id
        self.ec2_settings['role_instance_profile'] = role_instance_profile
        self.ec2_settings['master_elastic_ip'] = master_elastic_ip
        self.ec2_settings['bastion_node_public_dns'] = bastion_node_public_dns

    def set_hadoop_settings(self, master_instance_type=None,
                            slave_instance_type=None,
                            initial_nodes=None,
                            max_nodes=None,
                            custom_config=None,
                            slave_request_type=None,
                            use_hbase=None,
                            custom_ec2_tags=None,
                            use_hadoop2=None,
                            use_spark=None,
                            is_ha=None):
        """
        Kwargs:

        `master_instance_type`: The instance type to use for the Hadoop master
            node.

        `slave_instance_type`: The instance type to use for the Hadoop slave
            nodes.

        `initial_nodes`: Number of nodes to start the cluster with.

        `max_nodes`: Maximum number of nodes the cluster may be auto-scaled up
            to.

        `custom_config`: Custom Hadoop configuration overrides.

        `slave_request_type`: Purchasing option for slave instances.
            Valid values: "ondemand", "hybrid", "spot".

        `use_hbase`: Start hbase daemons on the cluster. Uses Hadoop2

        `use_hadoop2`: Use hadoop2 in this cluster

        `use_spark`: Use spark in this cluster

        `is_ha` : enable HA config for cluster

        """
        self.hadoop_settings['master_instance_type'] = master_instance_type
        self.hadoop_settings['slave_instance_type'] = slave_instance_type
        self.hadoop_settings['initial_nodes'] = initial_nodes
        self.hadoop_settings['max_nodes'] = max_nodes
        self.hadoop_settings['custom_config'] = custom_config
        self.hadoop_settings['slave_request_type'] = slave_request_type
        self.hadoop_settings['use_hbase'] = use_hbase
        self.hadoop_settings['use_hadoop2'] = use_hadoop2
        self.hadoop_settings['use_spark'] = use_spark
        self.hadoop_settings['is_ha'] = is_ha

        if custom_ec2_tags and custom_ec2_tags.strip():
            try:
                self.hadoop_settings['custom_ec2_tags'] = json.loads(custom_ec2_tags.strip())
            except Exception as e:
                raise Exception("Invalid JSON string for custom ec2 tags: %s" % e.message)

    def set_spot_instance_settings(self, maximum_bid_price_percentage=None,
                                   timeout_for_request=None,
                                   maximum_spot_instance_percentage=None):
        """
        Purchase options for spot instances. Valid only when
        `slave_request_type` is hybrid or spot.

        `maximum_bid_price_percentage`: Maximum value to bid for spot
            instances, expressed as a percentage of the base price for the
            slave node instance type.

        `timeout_for_request`: Timeout for a spot instance request (Unit:
            minutes)

        `maximum_spot_instance_percentage`: Maximum percentage of instances
            that may be purchased from the AWS Spot market. Valid only when
            slave_request_type is "hybrid".
        """
        self.hadoop_settings['spot_instance_settings'] = {
               'maximum_bid_price_percentage': maximum_bid_price_percentage,
               'timeout_for_request': timeout_for_request,
               'maximum_spot_instance_percentage': maximum_spot_instance_percentage}


    def set_stable_spot_instance_settings(self, maximum_bid_price_percentage=None,
                                          timeout_for_request=None,
                                          allow_fallback=True):
        """
        Purchase options for stable spot instances.

        `maximum_bid_price_percentage`: Maximum value to bid for stable node spot
            instances, expressed as a percentage of the base price
            (applies to both master and slave nodes).

        `timeout_for_request`: Timeout for a stable node spot instance request (Unit:
            minutes)

        `allow_fallback`: Whether to fallback to on-demand instances for
            stable nodes if spot instances are not available
        """
        self.hadoop_settings['stable_spot_instance_settings'] = {
               'maximum_bid_price_percentage': maximum_bid_price_percentage,
               'timeout_for_request': timeout_for_request,
               'allow_fallback': allow_fallback}


    def set_fairscheduler_settings(self, fairscheduler_config_xml=None,
                                   default_pool=None):
        """
        Fair scheduler configuration options.

        `fairscheduler_config_xml`: XML string with custom configuration
            parameters for the fair scheduler.

        `default_pool`: The default pool for the fair scheduler.
        """
        self.hadoop_settings['fairscheduler_settings'] = {
               'fairscheduler_config_xml': fairscheduler_config_xml,
               'default_pool': default_pool}

    def set_security_settings(self,
                              encrypted_ephemerals=None,
                              customer_ssh_key=None,
                              persistent_security_group=None):
        """
        Kwargs:

        `encrypted_ephemerals`: Encrypt the ephemeral drives on the instance.

        `customer_ssh_key`: SSH key to use to login to the instances.
        """
        self.security_settings['encrypted_ephemerals'] = encrypted_ephemerals
        self.security_settings['customer_ssh_key'] = customer_ssh_key
        self.security_settings['persistent_security_group'] = persistent_security_group

    def set_presto_settings(self, enable_presto=None, presto_custom_config=None):
        """
        Kwargs:

        `enable_presto`: Enable Presto on the cluster.

        `presto_custom_config`: Custom Presto configuration overrides.
        """
        self.presto_settings['enable_presto'] = enable_presto
        self.presto_settings['custom_config'] = presto_custom_config

    def minimal_payload(self):
        """
        This method can be used to create the payload which is sent while
        creating or updating a cluster.
        """
        payload = {"cluster": self.__dict__}
        return _make_minimal(payload)

class ClusterInfoV13():
    """
    qds_sdk.ClusterInfo is the class which stores information about a cluster.
    You can use objects of this class to create or update a cluster.
    """

    def __init__(self, label, api_version=1.3):
        """
        Args:

        `label`: A list of labels that identify the cluster. At least one label
            must be provided when creating a cluster.

        `api_version`: api version to use

        """
        self.label = label
        self.api_version = api_version
        self.ec2_settings = {}
        self.hadoop_settings = {}
        self.security_settings = {}
        self.presto_settings = {}
        self.node_configuration = {}

    def set_cluster_info(self, aws_access_key_id=None,
                         aws_secret_access_key=None,
                         aws_region=None,
                         aws_availability_zone=None,
                         vpc_id=None,
                         subnet_id=None,
                         master_elastic_ip=None,
                         disallow_cluster_termination=None,
                         enable_ganglia_monitoring=None,
                         node_bootstrap_file=None,
                         master_instance_type=None,
                         slave_instance_type=None,
                         initial_nodes=None,
                         max_nodes=None,
                         slave_request_type=None,
                         fallback_to_ondemand=None,
                         custom_config=None,
                         use_hbase=None,
                         custom_ec2_tags=None,
                         use_hadoop2=None,
                         use_spark=None,
                         use_qubole_placement_policy=None,
                         maximum_bid_price_percentage=None,
                         timeout_for_request=None,
                         maximum_spot_instance_percentage=None,
                         stable_maximum_bid_price_percentage=None,
                         stable_timeout_for_request=None,
                         stable_allow_fallback=True,
                         ebs_volume_count=None,
                         ebs_volume_type=None,
                         ebs_volume_size=None,
                         fairscheduler_config_xml=None,
                         default_pool=None,
                         encrypted_ephemerals=None,
                         ssh_public_key=None,
                         persistent_security_group=None,
                         enable_presto=None,
                         bastion_node_public_dns=None,
                         role_instance_profile=None,
                         presto_custom_config=None,
                         is_ha=None):
        """
        Kwargs:

        `aws_access_key_id`: The access key id for customer's aws account. This
            is required for creating the cluster.

        `aws_secret_access_key`: The secret access key for customer's aws
            account. This is required for creating the cluster.

        `aws_region`: AWS region to create the cluster in.

        `aws_availability_zone`: The availability zone to create the cluster
            in.

        `vpc_id`: The vpc to create the cluster in.

        `subnet_id`: The subnet to create the cluster in.

        `master_elastic_ip`: Elastic IP to attach to master node

        `disallow_cluster_termination`: Set this to True if you don't want
            qubole to auto-terminate idle clusters. Use this option with
            extreme caution.

        `enable_ganglia_monitoring`: Set this to True if you want to enable
            ganglia monitoring for the cluster.

        `node_bootstrap_file`: name of the node bootstrap file for this
            cluster. It should be in stored in S3 at
            <your-default-location>/scripts/hadoop/

        `master_instance_type`: The instance type to use for the Hadoop master
            node.

        `slave_instance_type`: The instance type to use for the Hadoop slave
            nodes.

        `initial_nodes`: Number of nodes to start the cluster with.

        `max_nodes`: Maximum number of nodes the cluster may be auto-scaled up
            to.

        `slave_request_type`: Purchasing option for slave instances.
            Valid values: "ondemand", "hybrid", "spot".

        `fallback_to_ondemand`: Fallback to on-demand nodes if spot nodes could not be
            obtained. Valid only if slave_request_type is 'spot'.

        `custom_config`: Custom Hadoop configuration overrides.

        `use_hbase`: Start hbase daemons on the cluster. Uses Hadoop2

        `use_hadoop2`: Use hadoop2 in this cluster

        `use_spark`: Use spark in this cluster

        `use_qubole_placement_policy`: Use Qubole Block Placement policy for 
            clusters with spot nodes.

        `maximum_bid_price_percentage`: ( Valid only when `slave_request_type` 
            is hybrid or spot.) Maximum value to bid for spot
            instances, expressed as a percentage of the base price 
            for the slave node instance type.

        `timeout_for_request`: Timeout for a spot instance request (Unit:
            minutes)

        `maximum_spot_instance_percentage`: Maximum percentage of instances
            that may be purchased from the AWS Spot market. Valid only when
            slave_request_type is "hybrid".

        `stable_maximum_bid_price_percentage`: Maximum value to bid for stable node spot
            instances, expressed as a percentage of the base price
            (applies to both master and slave nodes).

        `stable_timeout_for_request`: Timeout for a stable node spot instance request (Unit:
            minutes)

        `stable_allow_fallback`: Whether to fallback to on-demand instances for
            stable nodes if spot instances are not available

        `ebs_volume_count`: Number of EBS volumes to attach 
            to each instance of the cluster.

        `ebs_volume_type`: Type of the EBS volume. Valid 
            values are 'standard' (magnetic) and 'ssd'.

        `ebs_volume_size`: Size of each EBS volume, in GB. 

        `fairscheduler_config_xml`: XML string with custom configuration
            parameters for the fair scheduler.

        `default_pool`: The default pool for the fair scheduler.

        `encrypted_ephemerals`: Encrypt the ephemeral drives on the instance.

        `ssh_public_key`: SSH key to use to login to the instances.

        `persistent_security_group`: Comma-separated list of persistent 
            security groups for the cluster.

        `enable_presto`: Enable Presto on the cluster.

        `presto_custom_config`: Custom Presto configuration overrides.

        `bastion_node_public_dns`: Public dns name of the bastion node. Required only if cluster is in private subnet.

        `is_ha`: Enabling HA config for cluster

        """

        self.disallow_cluster_termination = disallow_cluster_termination
        self.enable_ganglia_monitoring = enable_ganglia_monitoring
        self.node_bootstrap_file = node_bootstrap_file
        self.set_node_configuration(master_instance_type, slave_instance_type, initial_nodes, max_nodes, slave_request_type, fallback_to_ondemand)
        self.set_ec2_settings(aws_access_key_id, aws_secret_access_key, aws_region, aws_availability_zone, vpc_id, subnet_id,
                              master_elastic_ip, bastion_node_public_dns, role_instance_profile)
        self.set_hadoop_settings(custom_config, use_hbase, custom_ec2_tags, use_hadoop2, use_spark, use_qubole_placement_policy, is_ha)
        self.set_spot_instance_settings(maximum_bid_price_percentage, timeout_for_request, maximum_spot_instance_percentage)
        self.set_stable_spot_instance_settings(stable_maximum_bid_price_percentage, stable_timeout_for_request, stable_allow_fallback)
        self.set_ebs_volume_settings(ebs_volume_count, ebs_volume_type, ebs_volume_size)
        self.set_fairscheduler_settings(fairscheduler_config_xml, default_pool)
        self.set_security_settings(encrypted_ephemerals, ssh_public_key, persistent_security_group)
        self.set_presto_settings(enable_presto, presto_custom_config)

    def set_ec2_settings(self,
                         aws_access_key_id=None,
                         aws_secret_access_key=None,
                         aws_region=None,
                         aws_availability_zone=None,
                         vpc_id=None,
                         subnet_id=None,
                         master_elastic_ip=None,
                         bastion_node_public_dns=None,
                         role_instance_profile=None):
        self.ec2_settings['compute_access_key'] = aws_access_key_id
        self.ec2_settings['compute_secret_key'] = aws_secret_access_key
        self.ec2_settings['aws_region'] = aws_region
        self.ec2_settings['aws_preferred_availability_zone'] = aws_availability_zone
        self.ec2_settings['vpc_id'] = vpc_id
        self.ec2_settings['subnet_id'] = subnet_id
        self.ec2_settings['master_elastic_ip'] = master_elastic_ip
        self.ec2_settings['bastion_node_public_dns'] = bastion_node_public_dns
        self.ec2_settings['role_instance_profile'] = role_instance_profile

    def set_node_configuration(self, master_instance_type=None,
                            slave_instance_type=None,
                            initial_nodes=None,
                            max_nodes=None,
                            slave_request_type=None,
                            fallback_to_ondemand=None):
        self.node_configuration['master_instance_type'] = master_instance_type
        self.node_configuration['slave_instance_type'] = slave_instance_type
        self.node_configuration['initial_nodes'] = initial_nodes
        self.node_configuration['max_nodes'] = max_nodes
        self.node_configuration['slave_request_type'] = slave_request_type
        self.node_configuration['fallback_to_ondemand'] = fallback_to_ondemand

    def set_hadoop_settings(self, custom_config=None,
                            use_hbase=None,
                            custom_ec2_tags=None,
                            use_hadoop2=None,
                            use_spark=None,
                            use_qubole_placement_policy=None,
                            is_ha=None):
        self.hadoop_settings['custom_config'] = custom_config
        self.hadoop_settings['use_hbase'] = use_hbase
        self.hadoop_settings['use_hadoop2'] = use_hadoop2
        self.hadoop_settings['use_spark'] = use_spark
        self.hadoop_settings['use_qubole_placement_policy'] = use_qubole_placement_policy
        self.hadoop_settings['is_ha'] = is_ha

        if custom_ec2_tags and custom_ec2_tags.strip():
            try:
                self.hadoop_settings['custom_ec2_tags'] = json.loads(custom_ec2_tags.strip())
            except Exception as e:
                raise Exception("Invalid JSON string for custom ec2 tags: %s" % e.message)

    def set_spot_instance_settings(self, maximum_bid_price_percentage=None,
                                   timeout_for_request=None,
                                   maximum_spot_instance_percentage=None):
        self.node_configuration['spot_instance_settings'] = {
               'maximum_bid_price_percentage': maximum_bid_price_percentage,
               'timeout_for_request': timeout_for_request,
               'maximum_spot_instance_percentage': maximum_spot_instance_percentage}

    def set_stable_spot_instance_settings(self, maximum_bid_price_percentage=None,
                                          timeout_for_request=None,
                                          allow_fallback=True):
        self.node_configuration['stable_spot_instance_settings'] = {
               'maximum_bid_price_percentage': maximum_bid_price_percentage,
               'timeout_for_request': timeout_for_request,
               'allow_fallback': allow_fallback}

    def set_ebs_volume_settings(self, ebs_volume_count=None,
                                 ebs_volume_type=None,
                                 ebs_volume_size=None):
      self.node_configuration['ebs_volume_count'] = ebs_volume_count
      self.node_configuration['ebs_volume_type'] = ebs_volume_type
      self.node_configuration['ebs_volume_size'] = ebs_volume_size


    def set_fairscheduler_settings(self, fairscheduler_config_xml=None,
                                   default_pool=None):
        self.hadoop_settings['fairscheduler_settings'] = {
               'fairscheduler_config_xml': fairscheduler_config_xml,
               'default_pool': default_pool}

    def set_security_settings(self,
                              encrypted_ephemerals=None,
                              ssh_public_key=None,
                              persistent_security_group=None):
        self.security_settings['encrypted_ephemerals'] = encrypted_ephemerals
        self.security_settings['ssh_public_key'] = ssh_public_key
        self.security_settings['persistent_security_group'] = persistent_security_group

    def set_presto_settings(self, enable_presto=None, presto_custom_config=None):
        self.presto_settings['enable_presto'] = enable_presto
        self.presto_settings['custom_config'] = presto_custom_config

    def minimal_payload(self):
        """
        This method can be used to create the payload which is sent while
        creating or updating a cluster.
        """
        payload_dict = self.__dict__
        payload_dict.pop("api_version", None)
        return _make_minimal(payload_dict)


def _make_minimal(dictionary):
    """
    This function removes all the keys whose value is either None or an empty
    dictionary.
    """
    new_dict = {}
    for key, value in dictionary.items():
        if value is not None:
            if isinstance(value, dict):
                new_value = _make_minimal(value)
                if new_value:
                    new_dict[key] = new_value
            else:
                new_dict[key] = value
    return new_dict



"""
The Template Module contains the base definition for Executing Templates
"""
import json
import logging
import os

from argparse import ArgumentParser
from qds_sdk.qubole import Qubole
from qds_sdk.resource import Resource

log = logging.getLogger("qds_template")

from qds_sdk.commands import *

class TemplateCmdLine:
    """
    qds_sdk.TemplateCmdLine is the interface used for template related operation in qds.py
    """
    
    @staticmethod
    def parsers():
        argparser = ArgumentParser(prog="qds.py template", description="Template Client for Qubole Data Service.")
        subparsers = argparser.add_subparsers()
        
        #create 
        create = subparsers.add_parser("create", help="To Create a new Template")
        create.add_argument("--data", dest="data",required=True, help="Path to JSON file with template details")
        create.set_defaults(func=TemplateCmdLine.create)
        
        #edit
        edit = subparsers.add_parser("edit", help="To Edit an existing Template")
        edit.add_argument("--data", dest="data", required=True, help="Path to JSON file with template details")
        edit.add_argument("--id", dest="id", required=True, help="Id for the Template")
        edit.set_defaults(func=TemplateCmdLine.edit)
        
        #clone
        clone = subparsers.add_parser("clone", help="To Clone an existing Template")
        clone.add_argument("--id", dest="id",required=True, help="Id for the Template to be Cloned")
        clone.add_argument("--data", dest="data", required=True, help="Path to JSON file with template details to override")
        clone.set_defaults(func=TemplateCmdLine.clone)
        
        #view
        view = subparsers.add_parser("view", help="To View an existing Template")
        view.add_argument("--id", dest="id", required=True, help="Id for the Template")
        view.set_defaults(func=TemplateCmdLine.view)
        
        #list 
        list = subparsers.add_parser("list", help="To List existing Templates")
        list.add_argument("--per-page", dest="per_page", help="Number of items per page")
        list.add_argument("--page", dest="page", help="Page Number")
        list.set_defaults(func=TemplateCmdLine.list)
        
        #run
        run = subparsers.add_parser("run", help="To Run Template and wait to print Result")
        run.add_argument("--id", dest="id", required=True, help="Id of the template to run")
        run.add_argument("--j", dest="data", required=True, help="Path to JSON file or json string with input field details")
        run.set_defaults(func=TemplateCmdLine.execute)
        
        #submit
        submit = subparsers.add_parser("submit", help="To Submit Template and get the command Id")
        submit.add_argument("--id", dest="id", required=True, help="Id of the template to Submit")
        submit.add_argument("--j", dest="data", required=True, help="Path to JSON file or json string with input field details")
        submit.set_defaults(func=TemplateCmdLine.submit)
        
        return argparser
        
    @staticmethod
    def run(args):      
        parser = TemplateCmdLine.parsers()
        parsed = parser.parse_args(args)
        return parsed.func(parsed)
    
    @staticmethod
    def create(args):
        with open(args.data) as f:
            spec = json.load(f)
        return Template.createTemplate(spec)
    
    @staticmethod
    def edit(args):
        with open(args.data) as f:
            spec = json.load(f)
        return Template.editTemplate(args.id, spec)
    
    @staticmethod
    def clone(args):
        with open(args.data) as f:
            spec = json.load(f)
        id = args.id
        return Template.cloneTemplate(id, spec)
        
    @staticmethod
    def submit(args):
        spec = getSpec(args)
        res = Template.submitTemplate(args.id, spec)
        log.info("Submitted Template  with Id: %s, Command Id: %s, CommandType: %s" % (args.id, res['id'], res['command_type']))
        return res
        
    @staticmethod    
    def execute(args):
        spec = getSpec(args)
        return Template.runTemplate(args.id, spec)
    
    @staticmethod
    def view(args):
        id = args.id
        return Template.viewTemplate(id)
    
    @staticmethod
    def list(args):
        return Template.listTemplates(args)    
    
def getSpec(args):
    if args.data is not None:
        if os.path.isfile(args.data):
            with open(args.data) as f:
                spec = json.load(f)
        else:
            spec = json.loads(args.data)
            if 'input_vars' in spec:
                inputs = formatData(spec['input_vars'])
                spec["input_vars"] = inputs
    else:
        spec = {}
    return spec


def formatData(inputs):
    res = []
    if len(inputs) != 0:
        for obj in inputs:
            o = {}
            for key in obj:
                o[key] = "'" + obj[key] + "'"
            res.append(o)
    return res
                    
class Template(Resource):
    """
    qds_sdk.Template is the base Qubole Template class.
    
    it uses /command_templates endpoint
    """
    
    rest_entity_path = "command_templates"
    
    @staticmethod
    def createTemplate(data):
        """
        Create a new template.

        Args:
            `data`: json data required for creating a template
        Returns:
            Dictionary containing the details of the template with its ID.
        """
        conn = Qubole.agent()
        return conn.post(Template.rest_entity_path, data)
    
    @staticmethod
    def editTemplate(id, data):
        """
        Edit an existing template.

        Args:
            `id`:   ID of the template to edit
            `data`: json data to be updated
        Returns:
            Dictionary containing the updated details of the template.
        """
        conn = Qubole.agent()
        return conn.put(Template.element_path(id), data)
    
    @staticmethod
    def cloneTemplate(id, data):
        """
        Clone an existing template.

        Args:
            `id`:   ID of the template to be cloned
            `data`: json data to override
        Returns:
            Dictionary containing the updated details of the template.
        """
        conn = Qubole.agent()
        path = str(id) + "/duplicate"
        return conn.post(Template.element_path(path), data)
        
    @staticmethod
    def viewTemplate(id):
        """
        View an existing Template details.

        Args:
            `id`: ID of the template to fetch
        
        Returns:
            Dictionary containing the details of the template.
        """
        conn = Qubole.agent()
        return conn.get(Template.element_path(id))
    
    @staticmethod
    def submitTemplate(id, data):
        """
        Submit an existing Template.

        Args:
            `id`: ID of the template to submit
            `data`: json data containing the input_vars 
        Returns:
            Dictionary containing Command Object details.  
        """
        conn = Qubole.agent()
        path = str(id) + "/run"
        return conn.post(Template.element_path(path), data)
    
    @staticmethod
    def runTemplate(id, data):
        """
        Run an existing Template and waits for the Result.
        Prints result to stdout. 

        Args:
            `id`: ID of the template to run
            `data`: json data containing the input_vars
        
        Returns:  
            An integer as status (0: success, 1: failure)
        """
        conn = Qubole.agent()
        path = str(id) + "/run"
        res = conn.post(Template.element_path(path), data)
        cmdType = res['command_type']
        cmdId = res['id']
        cmdClass = eval(cmdType)
        cmd = cmdClass.find(cmdId)
        while not Command.is_done(cmd.status):
            time.sleep(Qubole.poll_interval)
            cmd = cmdClass.find(cmd.id)
        return Template.getResult(cmdClass, cmd)
    
    @staticmethod
    def getResult(cmdClass, cmd):
        if Command.is_success(cmd.status):
            log.info("Fetching results for %s, Id: %s" % (cmdClass.__name__, cmd.id))
            cmd.get_results(sys.stdout, delim='\t')
            return 0
        else:
            log.error("Cannot fetch results - command Id: %s failed with status: %s" % (cmd.id, cmd.status))
            return 1
    
    @staticmethod
    def listTemplates(args):
        """
        Fetch existing Templates details.
        
        Args:
            `args`: dictionary containing the value of page number and per-page value
        Returns:
            Dictionary containing paging_info and command_templates details
        """
        conn = Qubole.agent()
        url_path = Template.rest_entity_path
        page_attr = []
        if args.page is not None:
            page_attr.append("page=%s" % args.page)
        if args.per_page is not None:
            page_attr.append("per_page=%s" % args.per_page)
        if page_attr:
            url_path = "%s?%s" % (url_path, "&".join(page_attr))

        return conn.get(url_path)


import json
import sys

from qds_sdk.qubole import Qubole
from qds_sdk.resource import Resource
from qds_sdk.cmd_line import CmdLine
from argparse import ArgumentParser


class NezhaCmdLine(CmdLine):
    """
    qds_sdk.NezhaCmdLine is the interface used by qds.py.
    """

    arg_to_classname = {'cubes': 'NezhaCube',
                        'measures': 'NezhaMeasure',
                        'partitions': 'NezhaPartition',
                        'dimensions': 'NezhaDimension',
                        'data_sources': 'NezhaDataSource',
                        'default_datasource': 'NezhaDefaultDS'}
    classname = ""

    @staticmethod
    def parsers():
        argparser = ArgumentParser(
            prog="qds.py nezha",
            description="Nezha client for Qubole Data Service.")
        subparsers = argparser.add_subparsers()

        # Create Nezha-Entities
        create = subparsers.add_parser(
            "create", help="Create a new cube/partition/data_source")
        create.add_argument(
            "nezha_entity", choices=['cubes', 'partitions', 'data_sources'],
            help="Name of nezha-entity")
        create.add_argument(
            "--data", dest="data", required=True,
            help="Attributes of nezha-entity as a json object")
        create.set_defaults(func=NezhaCmdLine.create)

        # List Nezha-Entities
        list = subparsers.add_parser(
            "list",
            help="List cubes/partitions/data_sources/default-datasource")
        list.add_argument(
            "nezha_entity",
            choices=[
                'cubes', 'partitions', 'data_sources', 'default_datasource'],
            help="Name of nezha-entity")
        list.add_argument(
            "--fields", nargs="*", dest="fields",
            help="List of fields to show for nezha-entity")
        list.add_argument(
            "--per-page", dest="per_page",
            help="Number of items per page")
        list.add_argument(
            "--page", dest="page",
            help="Page Number")
        list.set_defaults(func=NezhaCmdLine.list)

        # View Nezha-Entites
        view = subparsers.add_parser(
            "view", help="View a specific cube/partition/data_source")
        view.add_argument(
            "nezha_entity", choices=['cubes', 'partitions', 'data_sources'],
            help="Name of nezha-entity")
        view.add_argument(
            "id", help="Numeric id for entity")
        view.add_argument(
            "--fields", nargs="*", dest="fields",
            help="List of fields to show for nezha-entity")
        view.set_defaults(func=NezhaCmdLine.view)

        # Update Nezha-Entites
        update = subparsers.add_parser(
            "update",
            help="Update cube/partition/data_source/default_datasource")
        NezhaCmdLine._update_parsers(update)

        # Delete Nezha-Entites
        delete = subparsers.add_parser(
            "delete", help="Delete a cube/partition/data_source")
        delete.add_argument(
            "nezha_entity", choices=['cubes', 'partitions', 'data_sources'],
            help="Name of nezha-entity")
        delete.add_argument(
            "id", help="Numeric id of the nezha-entity to be deleted")
        delete.set_defaults(func=NezhaCmdLine.delete)

        return argparser

    @staticmethod
    def _update_parsers(update):
        update_subparsers = update.add_subparsers(
            help="update cubes/partitions/data_sources/default_datasource",
            dest="nezha_entity")

        parser_datasource = update_subparsers.add_parser(
            "data_sources", help="Update data-sources")
        parser_datasource.add_argument(
            "id", help="Numeric id of the DataSource")
        parser_datasource.add_argument(
            "--url", dest="url", help="URL of the data source")
        parser_datasource.add_argument(
            "--type", dest="type", help="Type of data source")
        parser_datasource.add_argument(
            "--name", dest="name", help="Name of the data source")
        parser_datasource.add_argument(
            "--user", dest="user", help="Username for data source")
        parser_datasource.add_argument(
            "--password", dest="password", help="Password for data source")
        parser_datasource.add_argument(
            "--auth_token", dest="auth_token",
            help="Authentication token for qds")
        parser_datasource.add_argument(
            "--dbtap_id", dest="dbtap_id", help="Id of the data-source in qds")

        parser_cube = update_subparsers.add_parser(
            "cubes", help="update cubes")
        parser_cube.add_argument(
            "id", help="Numeric id of the Cube")
        parser_cube.add_argument(
            "--cost", dest="cost", help="Cost for cube")
        parser_cube.add_argument(
            "--description", dest="description",
            help="Description of the cube")
        parser_cube.add_argument(
            "--name", dest="name", help="Name of the Cube")
        parser_cube.add_argument(
            "--schema_name", dest="schema_name", help="Schema name")
        parser_cube.add_argument(
            "--table_name", dest="table_name", help="Table name")
        parser_cube.add_argument(
            "--grouping_column", dest="grouping_column",
            help="Grouping Column")
        parser_cube.add_argument(
            "--destination_id", dest="destination_id",
            help="Data Source id acting as destiantion")
        parser_cube.add_argument("--query", dest="query", help="Query")

        parser_partition = update_subparsers.add_parser(
            "partitions", help="update partitions")
        parser_partition.add_argument(
            "id", help="Numeric id of the Partition")
        parser_partition.add_argument(
            "--cost", dest="cost", help="Cost for partition")
        parser_partition.add_argument(
            "--description", dest="description",
            help="Description of the partition")
        parser_partition.add_argument(
            "--name", dest="name", help="Name of the Partition")
        parser_partition.add_argument(
            "--schema_name", dest="schema_name", help="Schema name")
        parser_partition.add_argument(
            "--table_name", dest="table_name", help="Table name")
        parser_partition.add_argument(
            "--destination_id", dest="destination_id", help="Data Source id")
        parser_partition.add_argument("--query", dest="query", help="Query")

        parser_ds = update_subparsers.add_parser(
            "default_datasource", help="update default-datasource")
        parser_ds.add_argument(
            "default_datasource_id", help="DataSource id which is to be made default-datasource")

        update.set_defaults(func=NezhaCmdLine.update, default=None)

    @staticmethod
    def run(args):
        parser = NezhaCmdLine.parsers()
        parsed = parser.parse_args(args)
        NezhaCmdLine.classname = NezhaCmdLine.arg_to_classname[
            parsed.nezha_entity]
        return parsed.func(globals()[NezhaCmdLine.classname], parsed)

    @staticmethod
    def update(cls, args):
        args_dict = {}
        for key, value in args.__dict__.items():
            params = (key != "func" and key != "id" and key != "nezha_entity")
            if params and value is not None:
                args_dict[key] = value

        if args.nezha_entity == "default_datasource":
            return json.dumps(
                globals()[NezhaCmdLine.classname].update(**args_dict),
                sort_keys=True, indent=4)

        return json.dumps(
            globals()[NezhaCmdLine.classname].update(args.id, **args_dict),
            sort_keys=True, indent=4)


class NezhaDataSource(Resource):
    """
    qds_sdk.NezhaDataSource: is the class for retrieving and
    manipulating Qubole's NezhaDataSource.

    Endpoint for data source is: /nezha_data_sources.
    """

    rest_entity_path = "nezha_data_sources"


class NezhaPartition(Resource):
    """
    qds_sdk.NezhaPartition: is the class for retrieving and
    manipulating Qubole's NezhaPartition.

    Endpoint for partitions is: /nezha_partitions.
    """

    rest_entity_path = "nezha_partitions"


class NezhaCube(Resource):
    """
    qds_sdk.NezhaCube: is the class for retrieving and
    manipulating Qubole's NezhaCube.

    Endpoint for cubes is: /nezha_cubes.
    """

    rest_entity_path = "nezha_cubes"


class NezhaDimension(Resource):
    """
    qds_sdk.NezhaDimension: is the class for retrieving and
    manipulating Qubole's NezhaDimension.

    Endpoint for dimensions is: /nezha_cubes/{nezha_cube_id}/nezha_dimensions.
    User needs to set it before using.
    """

    rest_entity_path = ""


class NezhaMeasure(Resource):
    """
    qds_sdk.NezhaMeasure: is the class for retrieving and
    manipulating Qubole's NezhaMeasure.

    Endpoint for dimensions is: /nezha_cubes/{nezha_cube_id}/nezha_dimensions.
    User needs to set it before using.
    """

    rest_entity_path = ""


class NezhaDefaultDS(Resource):
    """
    qds_sdk.NezhaDefaultDS: is the class to update/get
    default datasource for NezhaDSSet class.
    """

    rest_entity_path = "accounts/default_datasource"

    @classmethod
    def update(cls, **kwargs):
        conn = Qubole.agent()
        return conn.put(cls.rest_entity_path, data=kwargs)



"""
The app module contains methods for managing apps.
"""
from qds_sdk.qubole import Qubole
from qds_sdk.resource import Resource
import argparse
import json


class AppCmdLine:
    """
    qds_sdk.AppCmdLine is the interface used by qds.py.
    """

    @staticmethod
    def parsers():
        argparser = argparse.ArgumentParser(
            prog="qds.py app",
            description="Client for managing apps on Qubole Data Service.")
        subparsers = argparser.add_subparsers(title="app operations")

        # Show information about an app
        show = subparsers.add_parser(
            "show", help="Show information about an app")
        show.add_argument("id", type=int, help="Numeric id of the app")
        show.set_defaults(func=AppCmdLine.show)

        # For listing information about all apps
        index = subparsers.add_parser(
            "list", help="List all available apps")
        index.set_defaults(func=AppCmdLine.index)

        def check_pair(pair):
            try:
                key, value = pair.split("=")
            except Exception:
                raise argparse.ArgumentTypeError(
                    "%s is an invalid key=value pair." % pair)
            return key, value

        # Create a new app
        create = subparsers.add_parser("create", help="Create a new app")
        create.add_argument("--name", required=True,
                            help="The name for the app")
        create.add_argument("--kind", default="spark", choices=['spark'],
                            help="The kind of the app. Default is spark")
        create.add_argument(
            "--config", nargs="*", type=check_pair,
            help="""Specify the config you want to set for this app as
            key=value pairs separated by spaces""")
        create.set_defaults(func=AppCmdLine.create)

        # Stop an app
        stop = subparsers.add_parser("stop", help="Stop an app")
        stop.add_argument("id", type=int, help="Numeric id of the app")
        stop.set_defaults(func=AppCmdLine.stop)

        # Delete an app
        delete = subparsers.add_parser("delete", help="Delete an app")
        delete.add_argument("id", type=int, help="Numeric id of the app")
        delete.set_defaults(func=AppCmdLine.delete)

        return argparser

    @staticmethod
    def run(args):
        parser = AppCmdLine.parsers()
        parsed = parser.parse_args(args)
        return parsed.func(parsed)

    @staticmethod
    def show(args):
        app = App.show(args.id)
        return json.dumps(app, indent=4)

    @staticmethod
    def index(args):
        result = App.index()
        return json.dumps(result, indent=4)

    @staticmethod
    def create(args):
        config = {}
        if args.config is not None:
            for key, value in args.config:
                config[key] = value
        result = App.create(name=args.name, config=config, kind=args.kind)
        return json.dumps(result, indent=4)

    @staticmethod
    def stop(args):
        result = App.stop(args.id)
        return json.dumps(result, indent=4)

    @staticmethod
    def delete(args):
        result = App.delete(args.id)
        return json.dumps(result, indent=4)


class App(Resource):
    """
    qds_sdk.App is the base Qubole Apps class.
    """

    """all apps use the /apps endpoint"""
    rest_entity_path = "apps"

    @classmethod
    def show(cls, app_id):
        """
        Shows an app by issuing a GET request to the /apps/ID endpoint.
        """
        conn = Qubole.agent()
        return conn.get(cls.element_path(app_id))

    @classmethod
    def index(cls):
        """
        Shows a list of all available apps by issuing a GET request to the
        /apps endpoint.
        """
        conn = Qubole.agent()
        return conn.get(cls.rest_entity_path)

    @classmethod
    def create(cls, name, config=None, kind="spark"):
        """
        Create a new app.

        Args:
            `name`: the name of the app

            `config`: a dictionary of key-value pairs

            `kind`: kind of the app (default=spark)
        """
        conn = Qubole.agent()
        return conn.post(cls.rest_entity_path,
                         data={'name': name, 'config': config, 'kind': kind})

    @classmethod
    def stop(cls, app_id):
        """
        Stops an app by issuing a PUT request to the /apps/ID/stop endpoint.
        """
        conn = Qubole.agent()
        return conn.put(cls.element_path(app_id) + "/stop")

    @classmethod
    def delete(cls, app_id):
        """
        Delete an app by issuing a DELETE request to the /apps/ID endpoint.
        """
        conn = Qubole.agent()
        return conn.delete(cls.element_path(app_id))
